"""Various plotting functions mainly based around plotly."""

import matplotlib.colors as mcolors
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import plotly.io as pio
import simplekml
from plotly.subplots import make_subplots

from . import processing

pio.templates["default"] = go.layout.Template(
    layout=go.Layout(
        margin=go.layout.Margin(l=0, r=0, b=0, t=0, pad=0),
    ),
)

pio.templates.default = "simple_white+default"


styling = {
    "colorscale": "geyser",
}


def blank_figure():
    fig = go.Figure()
    return fig


def scatter_3d(
    df: pd.DataFrame,
    color: str = "",
    colorbar_title: str = "",
    x: str = "utm_easting",
    y: str = "utm_northing",
    z: str = "altitude_ato",
    headings: bool = False,
):
    fig = px.scatter_3d(df, x=x, y=y, z=z)

    if color:
        custom_data = [df["timestamp"]]
        hovertemplate = [
            "northing: %{x:.2f}",
            "easting: %{y:.2f}",
            "altitude_ato: %{z:.2f}",
            f"{color}: %{{marker.color:.2f}}",
            "Time: %{customdata[0]}",
            "idx: %{pointNumber}",
        ]

        if headings:
            custom_data.extend([df["elevation_heading"], df["azimuth_heading"]])
            hovertemplate.extend(
                [
                    "elevation heading: %{customdata[1]:.2f}",
                    "azimuth heading: %{customdata[2]:.2f}",
                ]
            )

        fig.update_traces(
            marker=dict(
                color=df[color],
                size=4,
                opacity=0.5,
                colorscale=styling["colorscale"],
                colorbar=dict(title=colorbar_title),
            ),
            customdata=custom_data,
            hovertemplate="<br>".join(hovertemplate),
        )

    return fig


def scatter_2d(
    df: pd.DataFrame,
    x: str,
    color: str,
    y: str = "altitude_ato",
    **kwargs,
):
    fig = px.scatter(
        df,
        x=x,
        y=y,
        color=color,
        color_continuous_scale=styling["colorscale"],
        opacity=0.8,
        **kwargs,
    )
    fig.update_traces(
        customdata=df.index,
        hovertemplate="<br>".join(
            [
                "x: %{x:.2f}",
                "altitude_ato: %{y:.2f}",
                "CH4: %{marker.color:.2f}",
                "Time: %{customdata}",
            ],
        ),
    )

    return fig


def time_series(df: pd.DataFrame, y: str, y2=None, color=None, split=None):
    fig = go.Figure()
    fig.add_trace(
        go.Scatter(
            x=df.index,
            y=df[y],
            name=y,
            mode="markers",
        ),
    )
    if color is not None:
        fig.update_traces(marker_color=df[color], color_continuous_scale=styling["colorscale"])
    fig.update_traces(marker_size=8)
    if y2 is not None:
        fig.add_trace(
            go.Scatter(
                x=df.index,
                y=df[y2],
                name=y2,
                yaxis="y2",
                mode="markers",
            ),
        )
        fig.update_layout(yaxis2=dict(overlaying="y", side="right"))
    if split is not None:
        y_min, y_max = df[y].min(), df[y].max()
        fig.update_layout(
            shapes=[
                dict(
                    type="line",
                    xref="x",
                    yref="y",
                    x0=split,
                    y0=y_min,
                    x1=split,
                    y1=y_max,
                    line=dict(color="red", width=2),
                ),
            ],
        )
    fig.update(layout_yaxis_range=[0, df[y].max() * 1.05])
    fig.update_traces(opacity=0.5)
    return fig


def background_plotting(df: pd.DataFrame, y: str, bkg: np.ndarray, signal: pd.Series):
    fig = make_subplots(specs=[[{"secondary_y": True}]])
    ymin = df[y].min()
    ymax = df[y].max()
    ylim = [ymin * 0.95, ymax * 1.05]
    y2min = (df[y] - bkg).min()
    y2lim = (y2min, y2min + (ylim[1] - ylim[0]))
    df["normalised"] = df[y] - bkg
    fig.update_yaxes(range=ylim, secondary_y=False, title_text="Sensor CH4 (ppm)")
    fig.update_yaxes(range=y2lim, secondary_y=True, title_text="Normalised CH4 (ppm)")
    fig.add_scatter(x=df.index, y=df[y], opacity=0.3, name="Raw Data")
    fig.add_scatter(x=df.index, y=bkg, mode="lines", name="Background", line=dict(dash="dash"))
    fig.add_scatter(x=df.index, y=df["normalised"], yaxis="y2", name="Normalised Data", mode="lines", opacity=0.5)
    fig.add_scatter(x=signal.index, y=signal.values, yaxis="y2", name="Signal Points", mode="markers", opacity=0.5)

    return fig


def windrose_process(df: pd.DataFrame):
    beaufort = {
        "0": [0, 1],
        "1": [1, 2],
        "2": [2, 4],
        "3": [4, 6],
        "4": [6, 9],
        "5": [9, 11],
        "6": [11, 14],
        "7": [14, 17],
        "8": [17, 21],
        "9": [21, 25],
        "10": [25, 29],
        "11": [29, 33],
        "12": [33, 200],
    }

    beaufort_ms = {
        "0": "0-1",
        "1": "1-2",
        "2": "2-4",
        "3": "4-6",
        "4": "6-9",
        "5": "9-11",
        "6": "11-14",
        "7": "14-17",
        "8": "17-21",
        "9": "21-25",
        "10": "25-29",
        "11": "29-33",
        "12": "33+",
    }

    cardinals = {
        "N1": [0, 11.25],
        "NNE": [11.25, 33.75],
        "NE": [33.75, 56.25],
        "ENE": [56.25, 78.75],
        "E": [78.75, 101.25],
        "ESE": [101.25, 123.75],
        "SE": [123.75, 146.25],
        "SSE": [146.25, 168.75],
        "S": [168.75, 191.25],
        "SSW": [191.25, 213.75],
        "SW": [213.75, 236.25],
        "WSW": [236.25, 258.75],
        "W": [258.75, 281.25],
        "WNW": [281.25, 303.75],
        "NW": [303.75, 326.25],
        "NNW": [326.25, 348.75],
        "N2": [348.75, 360],
    }
    df["wind_direction_bin"] = pd.cut(
        df["winddir"],
        bins=[lower for lower, upper in cardinals.values()] + [list(cardinals.values())[-1][1]],
        labels=[key for key in cardinals],
        right=False,
    )

    df["wind_direction_bin"] = (
        df["wind_direction_bin"].map(lambda x: "N" if x in ["N1", "N2"] else x).astype("category")
    )
    df["beaufort"] = pd.cut(
        df["windspeed"],
        bins=[lower for lower, upper in beaufort.values()] + [list(beaufort.values())[-1][1]],
        labels=[key for key in beaufort],
        right=False,
    )
    df["beaufort_ms"] = df["beaufort"].map(beaufort_ms)
    df_windrose = df.groupby(["wind_direction_bin", "beaufort"], observed=False).size().reset_index(name="count")
    df_windrose["frequency"] = df_windrose["count"] / df_windrose["count"].sum() * 100
    df_windrose["wind_direction_bin_degs"] = df_windrose["wind_direction_bin"].cat.rename_categories(
        {
            "N": 0,
            "NNE": 22.5,
            "NE": 45,
            "ENE": 67.5,
            "E": 90,
            "ESE": 112.5,
            "SE": 135,
            "SSE": 157.5,
            "S": 180,
            "SSW": 202.5,
            "SW": 225,
            "WSW": 247.5,
            "W": 270,
            "WNW": 292.5,
            "NW": 315,
            "NNW": 337.5,
        },
    )
    df_windrose["beaufort"] = df_windrose["beaufort"].astype(int)
    return df_windrose


def windrose_graph(df, plot_transect=False, theta1=None, theta2=None):
    n_colors = 13
    colors = px.colors.sample_colorscale("turbo", [n / (n_colors - 1) for n in range(n_colors)])
    fig = px.bar_polar(
        df,
        r="frequency",
        theta="wind_direction_bin_degs",
        color="beaufort",
        labels={
            "frequency": "Frequency (%)",
            "wind_direction_bin": "Direction",
            "beaufort": "Beaufort Scale",
        },
        color_discrete_map=colors,
    )
    fig.update_layout(polar=dict(radialaxis={"visible": False, "showticklabels": False}))
    fig.update_layout(
        polar=dict(
            angularaxis={
                "showgrid": False,
            },
        ),
    )

    fig.update_layout(polar_bargap=0)
    if plot_transect:
        max_freq = df.groupby("wind_direction_bin", observed=False)["frequency"].sum().max()
        fig.add_trace(
            go.Scatterpolar(
                r=[max_freq, max_freq],
                theta=[theta1, theta2],
                mode="lines",
                line=dict(color="black", width=2, dash="dash"),
                showlegend=False,
            ),
        )
    return fig


def windrose(df: pd.DataFrame, plot_transect=False):
    df_windrose = windrose_process(df)
    if plot_transect:
        theta1, theta2 = processing.bimodal_azimuth(df)
        fig = windrose_graph(df_windrose, plot_transect=plot_transect, theta1=theta1, theta2=theta2)
    else:
        fig = windrose_graph(df_windrose, plot_transect=plot_transect)
    return fig


def outliers(original_data: pd.Series, fence_high: float, fence_low: float):
    outliers = np.array(original_data > fence_high) | (original_data < fence_low)

    fig = make_subplots(rows=1, cols=2, shared_yaxes=True)
    fig.add_trace(px.strip(original_data, color=outliers).data[0], row=1, col=1)
    if sum(outliers) > 0:
        fig.add_trace(px.strip(original_data, color=outliers).data[1], row=1, col=1)
        fig.add_shape(
            go.layout.Shape(
                type="line",
                x0=-0.5,
                y0=fence_high,
                x1=0.5,
                y1=fence_high,
                line=dict(color="red", width=2),
            ),
            row=1,
            col=1,
        )
        fig.add_shape(
            go.layout.Shape(
                type="line",
                x0=-0.5,
                y0=fence_low,
                x1=0.5,
                y1=fence_low,
                line=dict(color="red", width=2),
            ),
            row=1,
            col=1,
        )
    fig.update_traces(offsetgroup=0)

    fig.add_trace(px.scatter(original_data, color=outliers).data[0], row=1, col=2)
    if sum(outliers) > 0:
        fig.add_trace(px.scatter(original_data, color=outliers).data[1], row=1, col=2)
    fig.update_layout(showlegend=False, yaxis_title="Windspeed (ms⁻¹)")

    return fig


def contour_krig(
    df: pd.DataFrame, gas: str, xx: np.ndarray, yy: np.ndarray, field: np.ndarray, x: str = "x", y: str = "z"
) -> go.Figure:
    fig = go.Figure()
    fig.add_trace(
        go.Scatter(
            x=df[x],
            y=df[y],
            mode="markers",
            marker={
                "color": df[f"{gas}_normalised"],
                "colorscale": styling["colorscale"],
                "showscale": True,
                "colorbar": {
                    "title": f"{gas} (ppm)",
                },
            },
            showlegend=False,
        ),
    )
    fig.add_trace(
        go.Contour(
            z=field.T,
            x=xx[:, 0],
            y=yy[0, :],
            contours={
                "start": field.min(),
                "end": field.max(),
                "size": field.max() / 21,
            },
            colorscale=styling["colorscale"],
            opacity=0.5,
            showlegend=False,
            showscale=False,
        ),
    )
    fig.update_xaxes(
        showline=True,
        linewidth=1,
        linecolor="black",
        title_text="horizontal distance on projected flux plane (m)",
        range=[xx.min(), xx.max()],
        ticks="outside",
        tickwidth=1,
        tickcolor="black",
        ticklen=5,
        nticks=20,
    )
    fig.update_yaxes(
        showline=True,
        linewidth=1,
        linecolor="black",
        title_text="height above ground level (m)",
        range=[yy.min(), yy.max()],
        ticks="outside",
        tickwidth=1,
        tickcolor="black",
        ticklen=5,
        nticks=10,
    )
    fig.layout.coloraxis.colorbar.title = "Emissions flux (kg⋅m⁻²⋅h⁻¹)"

    return fig


def heatmap_krig(xx: np.ndarray, yy: np.ndarray, field: np.ndarray):
    fig = px.imshow(field.T, x=xx[:, 0], y=yy[0, :], color_continuous_scale=styling["colorscale"], origin="lower")
    fig.layout.coloraxis.colorbar.title = "Emissions flux (kg⋅m⁻²⋅h⁻¹)"
    fig.update_xaxes(
        showline=True,
        linewidth=1,
        linecolor="black",
        title_text="horizontal distance on cylindrical projected flux plane (m)",
        range=[xx.min(), xx.max()],
        ticks="outside",
        tickwidth=1,
        tickcolor="black",
        ticklen=5,
        nticks=20,
    )
    fig.update_yaxes(
        showline=True,
        linewidth=1,
        linecolor="black",
        title_text="height above ground level (m)",
        range=[yy.min(), yy.max()],
        ticks="outside",
        tickwidth=1,
        tickcolor="black",
        ticklen=5,
        nticks=10,
    )
    fig.update_layout(coloraxis_colorbar=dict(len=0.25))
    return fig


def create_kml_file(data: pd.DataFrame, output_file: str, column: str, altitudemode: str):
    kml = simplekml.Kml()

    min_value = data[column].min()
    max_value = data[column].max()

    custom_colors = [
        "#008080",
        "#70a494",
        "#b4c8a8",
        "#f6edbd",
        "#edbb8a",
        "#de8a5a",
        "#ca562c",
    ]  # based on plotly geyser
    cmap = mcolors.LinearSegmentedColormap.from_list("custom_cmap", custom_colors)

    for _index, row in data.iterrows():
        col_normalized = (row[column] - min) / (max_value - min_value)
        color = mcolors.rgb2hex(cmap(col_normalized))

        pnt = kml.newpoint(coords=[(row["longitude"], row["latitude"], row["altitude_ato"])], altitudemode=altitudemode)
        pnt.iconstyle.icon.href = "http://maps.google.com/mapfiles/kml/shapes/placemark_circle.png"
        pnt.iconstyle.color = simplekml.Color.rgb(int(color[1:3], 16), int(color[3:5], 16), int(color[5:], 16))
        pnt.iconstyle.scale = 0.6
        pnt.description = f"Concentration: {row[column]} ppm"

    kml.save(output_file)



"""a script to split a folder of flights into multiple flights based on the monotonic changes in altitude.
Takes a csv file as an input and will split it out into several csvs.
Needs azimuth headings to work properly.
"""

import argparse
from pathlib import Path

import colorama
import numpy as np
import pandas as pd

import gasflux

colorama.init()


def main(target_dir, search_string, output_dir=None):
    df_list = {}
    for file in Path(target_dir).rglob(search_string):
        df = pd.read_csv(file)
        df_list[file] = df

    for path, df in df_list.items():
        print(colorama.Fore.WHITE + "----------------------------------------")
        df, groupdict = gasflux.processing.monotonic_transect_groups(df)
        last_transect = None
        last_group_trend = None
        for group in df["group"].unique():
            # leaving this out for now as the group folder is a good identifier
            # if len(df['group'].unique()) == 1:
            #     print(colorama.Fore.RED + f'Skipping{path.name} - only one group')
            #     continue
            group_df = df[df["group"] == group]
            avg_altitudes = group_df.groupby("transect")["altitude"].mean().values
            avg_change = (
                sum([avg_altitudes[i + 1] - avg_altitudes[i] for i in range(len(avg_altitudes) - 1)])
                / len(avg_altitudes)
                if len(avg_altitudes) > 1
                else 0
            )
            # what's the trend
            current_group_trend = "ascending" if avg_change > 0 else "descending"
            # add last transect from previous group if there's a change in trend
            if last_group_trend and last_group_trend != current_group_trend and last_transect is not None:
                group_df = pd.concat([last_transect, group_df])
                avg_altitudes = group_df.groupby("transect")["altitude"].mean().values
            avg_altitudes = np.array(avg_altitudes)
            # check if the group is monotonic
            is_monotonic = np.all(np.diff(avg_altitudes) > 0) or np.all(np.diff(avg_altitudes) < 0)
            if not is_monotonic:  # exception
                print(f"group {group} is not monotonic - check the code!")
            formatted_avg_altitudes = ", ".join([f"{alt:.1f}" for alt in avg_altitudes])
            # do it where transect is the maximum transect number
            last_transect = group_df[group_df["transect"] == group_df["transect"].max()].copy()
            last_transect.loc[:, "transect"] = 0
            last_group_trend = current_group_trend
            unique_transects = len(group_df["transect"].unique())
            if unique_transects < 3:
                print(
                    colorama.Fore.RED
                    + f"Not saving {group} - not enough transects ({unique_transects} transects \
                        at {formatted_avg_altitudes}m)"
                )
            else:
                path = Path(path)
                if output_dir:
                    output_path = Path(
                        Path(output_dir)
                        / f"{path.parents[1].name}"
                        / f"{path.parent.name}_{group}"
                        / f"{path.stem}_{group}.csv"
                    )
                else:
                    # assuming in date and time folder
                    output_path = Path(
                        path.parent.parent.parent.parent
                        / "splits"
                        / path.parent.parent.name
                        / f"{path.parent.name}_{group}"
                        / f"{path.stem}_{group}.csv"
                    )
                output_path.parent.mkdir(parents=True, exist_ok=True)
                group_df.to_csv(output_path, index=False)
                print(
                    colorama.Fore.GREEN
                    + f'wrote {unique_transects} monotonic transects at {formatted_avg_altitudes}m to \n'
                    f'{"/".join(output_path.parts[-5:])}'
                )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--target-dir", help="the directory to search for csvs", default="survey")
    parser.add_argument("--search-string", help="the string to search for in the directory", default="*filtered.csv")
    parser.add_argument("--output-dir", help="the metadirectory to save the date/time/csvs to", required=False)
    args = parser.parse_args()
    main(args.target_dir, args.search_string, args.output_dir)



"""Functions that organise the data into standard columns in pandas dataframes. Conversion functions (e.g. WGS84 to UTM)
are here but transformations take place in processing.py"""

import geopandas as gpd
import numpy as np
import pandas as pd

from . import plotting
from .processing import circ_median


def data_tests(df: pd.DataFrame):
    assert df["ch4"].min() > 1.6, "ch4 values are too low"
    assert df.index.is_monotonic_increasing, "data is not sorted by time"
    assert df.index.is_unique, "data has duplicate timestamps"
    assert df["ch4"].isna().sum() == 0, "ch4 has missing values"
    assert df["windspeed"].min() >= 0, "windspeed values are negative"
    assert df["windspeed"].max() < 20, "windspeed values are too high"
    if df["windspeed"].max() > 15:
        print("Warning: windspeed is greater than 15 m/s, perhaps due to errors in the data.")


# make timestamp column from UTCs, Month, Day, Year
def timestamp_from_four_columns(df):
    df["Year"] = df["Year"] + 2000
    df["time"] = pd.to_datetime(df["UTCs"], unit="s")
    df["date"] = pd.to_datetime(df[["Year", "Month", "Day"]])
    df["timestamp"] = pd.to_datetime(df["date"].dt.date.astype(str) + " " + df["time"].dt.time.astype(str))
    df.index = df["timestamp"]
    df.drop(
        ["Year", "Month", "Day", "time", "date", "timestamp", "UTCs"],
        axis=1,
        inplace=True,
    )
    return df


# add UTM from latitudes and longitudes
def add_utm(df: pd.DataFrame) -> pd.DataFrame:
    gdf = gpd.GeoDataFrame(  # type: ignore
        df,
        geometry=gpd.points_from_xy(df["longitude"], df["latitude"], crs="EPSG:4326"),
    )
    utm = gdf.estimate_utm_crs()
    gdf = gdf.to_crs(utm)
    if not isinstance(gdf, gpd.GeoDataFrame):
        raise TypeError("Failed to reproject to a GeoDataFrame")
    gdf["utm_easting"] = gdf.geometry.x  # type: ignore
    gdf["utm_northing"] = gdf.geometry.y  # type: ignore
    output_df = pd.DataFrame(gdf)
    return output_df


# add columns for drone azimuth and elevation headings
def add_heading(df, rolling_window=1):
    df["hor_distance"] = np.sqrt((df["utm_northing"].diff()) ** 2 + (df["utm_easting"].diff()) ** 2)
    df["vert_distance"] = df["altitude_ato"].diff()
    df["vert_distance"] = pd.to_numeric(df["vert_distance"], errors="coerce")
    df["hor_distance"] = pd.to_numeric(df["hor_distance"], errors="coerce")
    df["azimuth_heading"] = (
        (np.degrees(np.arctan2(df["utm_easting"].diff(), df["utm_northing"].diff())) % 360)
        .rolling(rolling_window)
        .apply(lambda x: circ_median(x), raw=True)
    )
    df["elevation_heading"] = (
        np.degrees(np.arctan2(df["vert_distance"], df["hor_distance"]))
        .rolling(rolling_window)
        .apply(lambda x: circ_median(x), raw=True)
    )
    return df


def manual_filtering(dict_dfs: dict, split_times: dict, mask_spans: dict) -> dict:
    filtered_dfs = {}
    for name, df in dict_dfs.items():
        if name in mask_spans:
            for i in range(len(mask_spans[name])):
                df = df.drop(
                    df.between_time(mask_spans[name][i].split(" - ")[0], mask_spans[name][i].split(" - ")[1]).index,
                ).copy()
                filtered_dfs[name] = df.copy()
        if name in split_times:
            split_times[name].append("23:59:59")
            split_times[name].insert(0, "00:00:00")
            for i in range(len(split_times[name]) - 1):
                df2 = df.between_time(split_times[name][i], split_times[name][i + 1]).copy()
                filtered_dfs[name + "_" + str(i)] = df2.copy()
        elif name not in split_times:
            filtered_dfs[name] = df.copy()
    return filtered_dfs


def remove_outliers(df: pd.DataFrame, column: str, name: str):
    q1 = df[column].quantile(0.25)
    q3 = df[column].quantile(0.75)
    iqr = q3 - q1
    fence_low = q1 - 3 * iqr
    fence_high = q3 + 3 * iqr
    fig = plotting.outliers(df[column], fence_high, fence_low)
    outliers = df.loc[(df[column] < fence_low) | (df[column] > fence_high)]
    if len(outliers) > 0:
        print(f"{len(outliers)} outliers removed from {name} {column} data")
        # nan for outliers, not row removal
        df.loc[(df[column] < fence_low) | (df[column] > fence_high), column] = float("nan")
    elif len(outliers) == 0:
        print(f"No outliers found in {name} {column} data")

    return df, fig



"""A script to generate minimum required input data for testing and running the gasflux package."""

import numpy as np
import matplotlib.pyplot as plt
import noise
import yaml
from pathlib import Path
import pandas as pd
from geopy.distance import geodesic
from geopy.point import Point
import gasflux


def load_config() -> dict:
    file_path = Path(__file__).parent / "testdata" / "testconfig.yaml"
    with open(file_path) as file:
        return yaml.safe_load(file)


params = load_config()

np.random.seed(params["seed"])  # type: ignore


class SimulatedData2D:
    def __init__(self) -> None:  #
        self.horizontal_pixels = params["horizontal_pixels"]
        self.vertical_pixels = params["vertical_pixels"]
        self.num_plumes = params["num_plumes"]
        self.groupiness = params["groupiness"]
        self.spread = params["spread"]
        self.windspeed_avg = params["windspeed_avg"]
        self.windspeed_rel_std = params["windspeed_rel_std"]
        self.surface_roughness = params["surface_roughness"]
        self.seed = params["seed"]
        self.simplex_octaves = params["simplex_octaves"]
        self.simplex_persistence = params["simplex_persistence"]
        self.simplex_lacunarity = params["simplex_lacunarity"]
        self.wind_reference_height = params["wind_reference_height"]
        self.winddir_avg = params["winddir_avg"]
        self.winddir_std = params["winddir_std"]
        self.timestamp = params["timestamp"]
        self.flight_time_seconds = params["flight_time_seconds"]
        self.sample_frequency = params["sample_frequency"]
        self.start_coords = params["start_coords"]
        self.transect_azimuth = params["transect_azimuth"]
        self.sampling_altitude_ato_range = params["sampling_altitude_ato_range"]
        self.sampling_horizontal_range = params["sampling_horizontal_range"]
        self.scene_altitude_range = params["scene_altitude_range"]
        self.scene_horizontal_range = params["scene_horizontal_range"]
        self.sampling_horizontal_range = params["sampling_horizontal_range"]
        self.number_of_transects = params["number_of_transects"]
        self.gases = params["gases"]
        self.temperature = params["temperature"]
        self.pressure = params["pressure"]
        self.pixel_size_x = self.scene_horizontal_range[1] / self.horizontal_pixels  # type: ignore
        self.pixel_size_y = self.scene_altitude_range[1] / self.vertical_pixels  # type: ignore
        self.concentration_maps: dict = {}
        self.total_points = self.flight_time_seconds * self.sample_frequency
        self.transect_length = self.scene_horizontal_range[1] - self.scene_horizontal_range[0]
        self.points_per_transect = self.total_points // self.number_of_transects

    def generate_windspeed_map(self):
        z0 = self.surface_roughness
        z_ref = self.wind_reference_height
        u_ref = np.random.normal(self.windspeed_avg, self.windspeed_rel_std)
        self.z = np.linspace(0.1, self.scene_altitude_range[1], self.vertical_pixels)
        self.u = u_ref * np.log(self.z / z0) / np.log(z_ref / z0)
        self.windspeed_map = np.tile(self.u[:, np.newaxis], (1, self.horizontal_pixels))
        noise_map = np.zeros((self.vertical_pixels, self.horizontal_pixels))
        for i in range(self.vertical_pixels):
            for j in range(self.horizontal_pixels):
                x = i / self.vertical_pixels
                y = j / self.horizontal_pixels
                noise_val = noise.pnoise2(
                    x,
                    y,
                    octaves=self.simplex_octaves,
                    persistence=self.simplex_persistence,
                    lacunarity=self.simplex_lacunarity,
                )
                noise_map[i, j] = noise_val
        noise_map_std = np.std(noise_map)
        noise_map_mean = np.mean(noise_map)
        centered_noise_map = noise_map - noise_map_mean
        scaled_noise_map = centered_noise_map * (self.windspeed_rel_std / noise_map_std)
        shifted_noise_map = scaled_noise_map + 1
        self.windspeed_map *= shifted_noise_map

    def generate_winddir_map(self):
        self.winddir_map = np.full((self.vertical_pixels, self.horizontal_pixels), self.winddir_avg, dtype=float)
        noise_map = np.zeros((self.vertical_pixels, self.horizontal_pixels))
        for i in range(self.vertical_pixels):
            for j in range(self.horizontal_pixels):
                x = j / self.horizontal_pixels
                y = i / self.vertical_pixels
                noise_val = noise.pnoise2(
                    x,
                    y,
                    octaves=self.simplex_octaves,
                    persistence=self.simplex_persistence,
                    lacunarity=self.simplex_lacunarity,
                    base=0,
                )
                noise_map[i, j] = noise_val
        noise_map_std = np.std(noise_map)
        noise_map_mean = np.mean(noise_map)
        centered_noise_map = noise_map - noise_map_mean
        scaled_noise_map = centered_noise_map * (self.winddir_std / noise_map_std)
        self.winddir_map += scaled_noise_map
        self.winddir_map = (self.winddir_map + self.transect_azimuth + 90) % 360

    def generate_concentration_map(self, gas: str):
        self.concentration_maps[gas] = np.zeros((self.vertical_pixels, self.horizontal_pixels))
        max_concentration = self.gases[gas][1]
        min_concentration = self.gases[gas][0]
        for _ in range(self.num_plumes):
            x_mean = (
                np.random.rand() * self.horizontal_pixels * (1 - self.groupiness)
                + self.horizontal_pixels * self.groupiness / 2
            )
            y_mean = (
                np.random.rand() * self.vertical_pixels * (1 - self.groupiness)
                + self.vertical_pixels * self.groupiness / 2
            )
            x_std = np.random.rand() * self.horizontal_pixels * self.spread + self.horizontal_pixels * 0.05
            y_std = np.random.rand() * self.vertical_pixels * self.spread + self.vertical_pixels * 0.05
            concentration = np.random.uniform(min_concentration, max_concentration)
            x, y = np.meshgrid(np.arange(self.horizontal_pixels), np.arange(self.vertical_pixels))
            gaussian = concentration * np.exp(
                -((x - x_mean) ** 2 / (2 * x_std**2) + (y - y_mean) ** 2 / (2 * y_std**2))
            )
            self.concentration_maps[gas] += gaussian
            self.concentration_maps[gas] = (self.concentration_maps[gas] - np.min(self.concentration_maps[gas])) / (
                np.max(self.concentration_maps[gas]) - np.min(self.concentration_maps[gas])
            )
            self.concentration_maps[gas] = (
                self.concentration_maps[gas] * (max_concentration - min_concentration) + min_concentration
            )

    def generate_data(self) -> None:
        self.generate_windspeed_map()
        for gas in self.gases:
            self.generate_concentration_map(gas)
        self.generate_winddir_map()

    def generate_sampling_flight(self) -> None:
        self.end_coords = geodesic(meters=self.transect_length).destination(
            Point(self.start_coords), float(self.transect_azimuth)
        )
        self.df = pd.DataFrame(
            index=range(self.total_points),
            columns=[
                "timestamp",
                "latitude",
                "longitude",
                "x",
                "altitude_ato",
                "windspeed",
                "temperature",
                "pressure",
            ],
        )
        for gas in self.gases:
            self.df[gas] = np.zeros(self.total_points)
        self.df["timestamp"] = pd.date_range(
            start=self.timestamp, periods=self.total_points, freq=f"{1/self.sample_frequency}s"
        )

        def geopoints(self):
            start_point = Point(self.start_coords)
            points = [start_point]
            distance_per_step = self.transect_length / self.points_per_transect

            for _ in range(1, self.points_per_transect):
                # Use geodesic to find the next point a certain distance towards the azimuth
                next_point = geodesic(meters=distance_per_step).destination(points[-1], self.transect_azimuth)
                points.append(next_point)
            return points

        points = geopoints(self)
        x_increment = (self.sampling_horizontal_range[1] - self.sampling_horizontal_range[0]) / self.points_per_transect
        x_series = np.linspace(
            self.sampling_horizontal_range[0],
            self.sampling_horizontal_range[0] + x_increment * self.points_per_transect,
            self.points_per_transect,
        )
        latitudes = []
        longitudes = []
        xes = []
        for i in range(0, self.number_of_transects):
            if i % 2 == 0:
                latitudes.extend([point.latitude for point in points])
                longitudes.extend([point.longitude for point in points])
                xes.extend(x_series)
            else:  # invert to allow for proper azimuth switches
                latitudes.extend([point.latitude for point in points[::-1]])
                longitudes.extend([point.longitude for point in points[::-1]])
                xes.extend(x_series[::-1])
        self.df["latitude"] = latitudes
        self.df["longitude"] = longitudes
        self.df["x"] = xes
        altitude_step = (
            self.sampling_altitude_ato_range[1] - self.sampling_altitude_ato_range[0]
        ) / self.number_of_transects

        for i in range(self.number_of_transects):
            start_index = i * self.points_per_transect
            end_index = start_index + self.points_per_transect
            self.df.loc[start_index : end_index - 1, "altitude_ato"] = (
                self.sampling_altitude_ato_range[0] + i * altitude_step
            )
        self.df["temperature"] = self.temperature
        self.df["pressure"] = self.pressure

        def sample_data(self):
            self.df["norm_x"] = (self.df["x"] - self.scene_horizontal_range[0]) / (
                self.scene_horizontal_range[1] - self.scene_horizontal_range[0]
            )
            self.df["norm_altitude_ato"] = (self.df["altitude_ato"] - self.scene_altitude_range[0]) / (
                self.scene_altitude_range[1] - self.scene_altitude_range[0]
            )
            self.df["index_x"] = np.floor(self.df["norm_x"] * (self.horizontal_pixels - 1)).astype(int)
            self.df["index_y"] = np.floor(self.df["norm_altitude_ato"] * (self.vertical_pixels - 1)).astype(int)
            self.df["index_x"] = np.clip(self.df["index_x"], 0, self.horizontal_pixels - 1)
            self.df["index_y"] = np.clip(self.df["index_y"], 0, self.vertical_pixels - 1)
            self.df["windspeed"] = self.windspeed_map[self.df["index_y"], self.df["index_x"]]
            self.df["winddir"] = self.winddir_map[self.df["index_y"], self.df["index_x"]]
            for gas in self.gases:
                self.df[gas] = self.concentration_maps[gas][self.df["index_y"], self.df["index_x"]]

        sample_data(self)

        self.df = gasflux.pre_processing.add_utm(self.df)
        self.df = gasflux.pre_processing.add_heading(self.df)

        self.df_min = self.df.copy()
        retained_columns = [
            "timestamp",
            "latitude",
            "longitude",
            "altitude_ato",
            "windspeed",
            "winddir",
            "temperature",
            "pressure",
        ]
        for gas in self.gases:
            retained_columns.append(gas)
        self.df_min = self.df_min[retained_columns]

    def plot_data(self, logwind: bool, windspeed: bool, winddir: bool, gas: bool, scatter_3d: bool) -> None:
        # Plot the log wind profile
        if logwind:
            fig, ax1 = plt.subplots(figsize=(6, 6))
            ax1.plot(self.z, self.u, "b-")
            ax1.set_ylabel("Wind Speed (m/s)")
            ax1.set_xlabel("Height (m)")
            ax1.set_title("Log Wind Profile")
            ax1.grid(True)
            plt.show()

        # Plot the wind speed map
        if windspeed:
            fig, ax2 = plt.subplots(figsize=(20, 8))
            im = ax2.imshow(
                self.windspeed_map,
                cmap="viridis",
                origin="lower",
                extent=[0, self.scene_horizontal_range[1], 0, self.scene_altitude_range[1]],
            )
            ax2.set_title("Wind Speed Map")
            ax2.set_xlabel("X (m)")
            ax2.set_ylabel("Y (m)")
            cbar = fig.colorbar(im, ax=ax2)
            cbar.set_label("Wind Speed (m/s)")
            plt.show()

        # Plot the concentration map(s)
        if gas:
            num_gases = len(self.gases)
            fig, axes = plt.subplots(num_gases, 1, figsize=(20, 6))
            fig.suptitle("Gas Concentration Maps", fontsize=16)

            for i, gas_name in enumerate(self.gases):
                im = axes[i].imshow(
                    self.concentration_maps[gas_name],
                    cmap="viridis",
                    origin="lower",
                    extent=[0, self.scene_horizontal_range[1], 0, self.scene_altitude_range[1]],
                )
                axes[i].set_title(f"{gas_name.upper()}")
                axes[i].set_xlabel("X (m)")
                axes[i].set_ylabel("Y (m)")
                cbar = fig.colorbar(im, ax=axes[i])
                cbar.set_label(f"{gas_name.upper()} Concentration (ppm)")

            plt.tight_layout()
            plt.show()

        # Plot the wind direction map
        if winddir:
            fig, ax4 = plt.subplots(figsize=(20, 8))
            im = ax4.imshow(
                self.winddir_map,
                cmap="hsv",
                origin="lower",
                extent=[0, self.scene_horizontal_range[1], 0, self.scene_altitude_range[1]],
            )
            ax4.set_title("Wind Direction Map")
            ax4.set_xlabel("X (m)")
            ax4.set_ylabel("Y (m)")
            cbar = fig.colorbar(im, ax=ax4)
            cbar.set_label("Wind Direction (Degrees)")
            plt.show()
        if scatter_3d:
            fig = gasflux.plotting.scatter_3d(self.df, color="ch4")
            fig.show()


def main():
    data = SimulatedData2D()
    data.generate_data()
    data.generate_sampling_flight()
    data.plot_data(logwind=False, windspeed=False, winddir=False, gas=False, scatter_3d=True)
    data.df_min.to_csv(Path(__file__).parent / "testdata" / "exampledata.csv", index=False)
    data.df.to_csv(Path(__file__).parent / "testdata" / "testdata.csv", index=False)


if __name__ == "__main__":
    main()



import pandas as pd
import gasflux
import yaml
from pathlib import Path
import numpy as np
from gasflux.processing import min_angular_displacement
import pytest


testdf = pd.read_csv(Path(__file__).parents[1] / "src" / "gasflux" / "testdata" / "testdata.csv")
testconfig = yaml.safe_load(open(Path(__file__).parents[1] / "src" / "gasflux" / "testdata" / "testconfig.yaml"))


def load_cols(cols):
    return testdf[cols]


def test_min_angular_diff_def():
    a = 0
    b = 359
    diff = gasflux.processing.min_angular_displacement(a, b)
    assert diff == 1, "Angular difference not calculated correctly"


def test_circ_median():
    x = np.array([0, 1, 2, 359, 4, 3])
    median = gasflux.processing.circ_median(x)
    assert median == 1.5, "Circular median not calculated correctly"


@pytest.mark.parametrize(
    "plane_angle,expected_winddir_rel,expected_windspeed_normal",
    [
        (
            90,
            [0, 90, 0, 90, 0],
            [5, 0, 5, 0, 5],
        ),
        (
            30,
            [60, 30, 60, 30, 60],
            np.array([1 / 2, np.sqrt(3) / 2, 1 / 2, np.sqrt(3) / 2, 1 / 2]) * 5,
        ),
        (
            60,
            [30, 60, 30, 60, 30],
            np.array([np.sqrt(3) / 2, 1 / 2, np.sqrt(3) / 2, 1 / 2, np.sqrt(3) / 2]) * 5,
        ),
    ],
)
def test_wind_offset_correction_parametrized(plane_angle, expected_winddir_rel, expected_windspeed_normal):
    data = {"winddir": [0, 90, 180, 270, 360], "windspeed": [5, 5, 5, 5, 5]}
    df = pd.DataFrame(data)
    corrected_df = gasflux.processing.wind_offset_correction(df, plane_angle)
    assert "winddir_rel" in corrected_df.columns, f"Relative wind direction column not added for angle {plane_angle}"
    assert "windspeed" in corrected_df.columns, f"Normalised wind speed column not added for angle {plane_angle}"
    assert np.allclose(
        corrected_df["winddir_rel"], expected_winddir_rel, rtol=1e-5, atol=1e-10
    ), f"Relative wind directions not calculated correctly for angle {plane_angle}"
    assert np.allclose(
        corrected_df["windspeed"], expected_windspeed_normal, rtol=1e-5, atol=1e-10
    ), f"Normalised wind speeds not calculated correctly for angle {plane_angle}"


def test_bimodal_azimuth():
    input_mode = testconfig["transect_azimuth"]
    input_reciprocal_mode = (input_mode + 180) % 360
    df = load_cols(["azimuth_heading", "altitude_ato"])
    mode1, mode2 = gasflux.processing.bimodal_azimuth(df)
    assert (
        min_angular_displacement(mode1, input_mode) < 3 or min_angular_displacement(mode1, input_reciprocal_mode) < 3
    ), "Mode1 does not match expected azimuth or its reciprocal within 3 degrees"

    if min_angular_displacement(mode1, input_mode) < 3:
        assert (
            min_angular_displacement(mode2, input_reciprocal_mode) < 3
        ), "Mode2 does not match expected reciprocal azimuth within 3 degrees"
    else:
        assert min_angular_displacement(mode2, input_mode) < 3, "Mode2 does not match expected azimuth within 3 degrees"


def test_bimodal_elevation():
    df = load_cols(["elevation_heading", "altitude_ato"])
    input_mode = 0
    input_reciprocal_mode = 0 - input_mode
    mode1, mode2 = gasflux.processing.bimodal_elevation(df)
    assert (
        min_angular_displacement(mode1, input_mode) < 3 or min_angular_displacement(mode1, input_reciprocal_mode) < 3
    ), "Mode1 does not match expected elevation or its reciprocal within 3 degrees"
    if min_angular_displacement(mode1, input_mode) < 3:
        assert (
            min_angular_displacement(mode2, input_reciprocal_mode) < 3
        ), "Mode2 does not match expected reciprocal elevation within 3 degrees"
    else:
        assert (
            min_angular_displacement(mode2, input_mode) < 3
        ), "Mode2 does not match expected elevation within 3 degrees"


def test_altitude_transect_splitter():
    df = load_cols(["altitude_ato"])
    df, fig = gasflux.processing.altitude_transect_splitter(df)
    assert "transect_num" in df.columns, "Transect number column not added to dataframe"
    assert (
        df["transect_num"].nunique() == testconfig["number_of_transects"]
    ), "Dataframe was not split into the right number of transects"


def test_add_transect_azimuth_switches():
    df = load_cols(["azimuth_heading"])
    df = gasflux.processing.add_transect_azimuth_switches(df)
    assert (
        df["transect_num"].nunique() == testconfig["number_of_transects"]
    ), "Transect azimuth switches not added to dataframe"


def test_heading_filter():
    df = load_cols(["azimuth_heading", "elevation_heading", "altitude_ato"])
    azimuth_filter = testconfig["filters"]["heading_filter"]["azimuth_filter"]
    azimuth_window = testconfig["filters"]["heading_filter"]["azimuth_window"]
    elevation_filter = testconfig["filters"]["heading_filter"]["elevation_filter"]
    df_filtered, df_unfiltered = gasflux.processing.heading_filter(
        df, azimuth_filter=azimuth_filter, azimuth_window=azimuth_window, elevation_filter=elevation_filter
    )
    input_mode = testconfig["transect_azimuth"]
    input_reciprocal_mode = (input_mode + 180) % 360
    # assert that the filtered dataframe contains the expected azimuth or its reciprocal within the window
    df_filtered["near_mode1"] = df_filtered["rolling_azimuth_heading"].apply(
        lambda x: min_angular_displacement(x, input_mode) < azimuth_window
    )
    df_filtered["near_mode2"] = df_filtered["rolling_azimuth_heading"].apply(
        lambda x: min_angular_displacement(x, input_reciprocal_mode) < azimuth_window
    )
    assert (
        df_filtered["near_mode1"].any() or df_filtered["near_mode2"].any()
    ), "Filtered dataframe does not contain expected azimuth or its reciprocal within the window"


def test_mCount_max():
    data_dict = {1: -5.4, 2: 0.6, 3: 5.6, 4: 3.2, 5: 10.4, 6: 18.4, 7: 20.8, 8: 19.4}
    start, end = gasflux.processing.mCount_max(data_dict)
    assert start == 4, "Start index of max count not calculated correctly"
    assert end == 7, "End index of max count not calculated correctly"


def test_largest_monotonic_transect_series():
    df = load_cols(
        ["timestamp", "altitude_ato", "azimuth_heading", "longitude", "latitude", "utm_easting", "utm_northing"]
    )
    df, starttransect, endtransect = gasflux.processing.largest_monotonic_transect_series(df)
    starttransect = 1
    endtransect = testconfig["number_of_transects"]
    assert starttransect == starttransect, "Start index of largest monotonic transect not calculated correctly"
    assert endtransect == endtransect, "End index of largest monotonic transect not calculated correctly"


def test_remove_non_transects():
    df = load_cols(
        ["altitude_ato", "azimuth_heading", "elevation_heading", "longitude", "latitude", "utm_easting", "utm_northing"]
    )
    retained_df, removed_df = gasflux.processing.remove_non_transects(df)
    assert retained_df is not None, "Retained dataframe is None"
    assert removed_df is not None, "Removed dataframe is None"


def test_flatten_linear_plane():
    df = load_cols(["altitude_ato", "utm_easting", "utm_northing"])
    df, plane_angle = gasflux.processing.flatten_linear_plane(df)
    plane_angle = np.degrees(plane_angle)
    input_plane_angle = testconfig["transect_azimuth"]
    reciprocal_plane_angle = (input_plane_angle + 180) % 360
    assert (
        min_angular_displacement(plane_angle, input_plane_angle) < 3
        or min_angular_displacement(plane_angle, reciprocal_plane_angle) < 3
    ), "Plane angle not calculated correctly"



"""Processing function, usually implying some kind of filtering or data transformation."""

from matplotlib.figure import Figure
from itertools import groupby
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import odr
from scipy.optimize import least_squares
from scipy.signal import find_peaks
from scipy.stats import circmean
import warnings
import logging


def circ_median(x: np.ndarray | pd.Series) -> float:
    """
    Finds the largest "empty" arc in the circle, splits it there and returns the median of the straight line.
    Probably there are more elegant ways... see https://github.com/scipy/scipy/issues/6644

    Parameters:
        x (np.ndarray | pd.Series): The input array of angles.

    Returns:
        float: The circular median of the input angles.
    """
    angles = sorted(x)
    n = len(angles)
    if n == 1:
        return angles[0]
    gaps = [(angles[(i + 1) % n] - angles[i]) % 360 for i in range(n)]
    largest_gap_index = gaps.index(max(gaps))
    rearranged_angles = angles[largest_gap_index + 1 :] + angles[: largest_gap_index + 1]
    if n % 2 == 1:
        return rearranged_angles[n // 2]
    else:
        return (rearranged_angles[n // 2 - 1] + rearranged_angles[n // 2]) % 360 / 2


def min_angular_displacement(x: float, y: float) -> float:
    """
    Calculates the minimum circular difference between two angles (in 360 degree space)
    """
    return min(abs(x - y) % 360, (360 - abs(x - y)) % 360)


def wind_offset_correction(df: pd.DataFrame, plane_angle: float) -> pd.DataFrame:
    """
    Corrects wind direction data for a given plane angle, assuming the plane is the primary orientation of the dataset.
    This is useful for aligning wind data with the plane's orientation, facilitating analysis of wind effects.

    Parameters:
        df (pd.DataFrame): The input dataframe containing wind direction data.
        wind_dir_col (str): Column name for wind direction data.
        plane_angle (float): Angle of the plane in degrees.

    Returns:
        pd.DataFrame: The modified dataframe with corrected wind direction data.
    """
    df = df.copy()
    df["winddir_rel"] = df.apply(lambda row: abs(90 - min_angular_displacement(row["winddir"], plane_angle)), axis=1)
    df["windspeed_measured"] = df["windspeed"]
    df["windspeed"] = df["windspeed"] * np.cos(np.radians(df["winddir_rel"]))
    return df


def bimodal_azimuth(
    df: pd.DataFrame, heading_col: str = "azimuth_heading", min_altitude: int = 5, min_diff: int = 160
) -> tuple[float, float]:
    """
    Identifies the two most frequent azimuth headings in the dataset, ensuring they are sufficiently
    distinct. Filters data by altitude and removes NaNs before analysis.

    Parameters:
        df (pd.DataFrame): The input dataframe.
        heading_col (str): Column name for azimuth heading data. Default is "azimuth_heading".
        min_altitude (int): Minimum altitude for data to be included. Default is 5.
        min_diff (int): Minimum difference between the two modes. Default is 160.

    Returns:
        tuple: Two modes of the azimuth heading.
    """
    df = df[df["altitude_ato"] >= min_altitude]
    data = df[heading_col].dropna().to_numpy()
    hist, edges = np.histogram(data, bins=50)
    edgedist = edges[1] - edges[0]
    bin_centers = edges[:-1] + edgedist / 2
    max_freq_idx = np.argsort(hist)[-2:]  #  top 2 frequencies
    mode1, mode2 = bin_centers[max_freq_idx]
    while np.abs(mode1 - mode2) < min_diff:
        if hist[max_freq_idx[0]] < hist[max_freq_idx[1]]:
            hist[max_freq_idx[0]] = 0
            max_freq_idx = np.argsort(hist)[-2:]
        else:
            hist[max_freq_idx[1]] = 0
            max_freq_idx = np.argsort(hist)[-2:]
        mode1, mode2 = bin_centers[max_freq_idx]
    if min_angular_displacement(mode1, mode2) < 160:
        warnings.warn(
            f"Two modes are close together - this probably should never happen: {mode1:.2f} and {mode2:.2f}",
            UserWarning,
            stacklevel=2,
        )
    return (mode1, mode2)


# this returns modes of slope from -90 to 90 degrees.
def bimodal_elevation(
    df: pd.DataFrame, heading_col: str = "elevation_heading", min_flight_altitude: float = 5, max_slope: float = 70
) -> tuple[float, float]:
    """
    Identifies the most frequent elevation heading in the dataset, adjusted for vertical movements.
    Filters data by altitude and removes NaNs before analysis.

    Parameters:
        df (pd.DataFrame): The input dataframe.
        heading_col (str): Column name for elevation heading data. Default is "elevation_heading".
        min_altitude (int): Minimum altitude for data to be included. Default is 5.
        max_slope (int): Maximum slope to consider, avoTupleiding vertical movements. Default is 70.

    Returns:
        tuple: Mode of elevation heading and its negative, representing possible ascent/descent angles.
    """
    df = df[df["altitude_ato"] >= df["altitude_ato"].min() + min_flight_altitude]
    data = df[heading_col].to_numpy()
    data = np.abs(data[~np.isnan(data)])
    # to get around the edge case where vertical movements are modal
    data = data[data < max_slope]
    hist, edges = np.histogram(data, bins=50)
    max_freq_idx = np.argsort(hist)[::-1][:2]
    mode = edges[max_freq_idx][0]
    return (mode, -mode)


def altitude_transect_splitter(df: pd.DataFrame) -> tuple[pd.DataFrame, Figure]:
    """
    Splits the dataset into altitude-based transects and plots histogram peaks to identify prominent
    altitude ranges. Only works if the flights are flat.

    Parameters:
        df (pd.DataFrame): The input dataframe containing altitude data.

    Returns:
        tuple: Modified dataframe with transect labels and a figure showing the histogram with peaks.
    """
    df = df.copy()
    altitudes = df["altitude_ato"].to_numpy()
    heights, bin_edges = np.histogram(altitudes, bins=40)
    bin_width = bin_edges[1] - bin_edges[0]
    bin_edges = np.append(altitudes.min() - bin_width, bin_edges)  # avoid literal edge effects
    bin_edges = np.append(bin_edges, altitudes.max() + bin_width)
    heights = np.append(0, heights)
    heights = np.append(heights, 0)
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2
    peaks, properties = find_peaks(heights)
    transect_edges = (bin_centers[peaks][:-1] + bin_centers[peaks][1:]) / 2
    transect_edges = np.append(altitudes.min(), transect_edges)
    transect_edges = np.append(transect_edges, altitudes.max())
    fig, ax = plt.subplots()
    ax.stairs(edges=bin_edges, values=heights, fill=True)
    ax.plot(bin_centers[peaks], heights[peaks], "x", color="red")
    ax.vlines(transect_edges, ymin=0, ymax=max(heights), color="red")
    df["transect_num"] = pd.cut(df["altitude_ato"], bins=list(transect_edges), labels=False, include_lowest=True)  # type: ignore
    return df, fig


def add_transect_azimuth_switches(df: pd.DataFrame, threshold=150, shift=3) -> pd.DataFrame:
    """
    Identifies transects based on significant changes in azimuth heading, incrementing a transect
    counter to distinguish different flight paths. This is a really crude function and should probably
    be only used with data that's already been filtered in some way.

    Parameters:
        df (pd.DataFrame): The input dataframe with azimuth headings.

    Returns:
        pd.DataFrame: The modified dataframe with a new 'transect' column indicating transect IDs.
    """
    df = df.copy()
    df["transect_num"] = 0
    df["prev_azimuth_heading"] = df["azimuth_heading"].shift(
        shift
    )  # this gives better behaviour for very neat transects
    df["deg_displace"] = df.apply(
        lambda row: min_angular_displacement(row["azimuth_heading"], row["prev_azimuth_heading"])
        if not pd.isnull(row["prev_azimuth_heading"])
        else np.nan,
        axis=1,
    )
    df.loc[df["deg_displace"] > threshold, "transect_num"] = 1
    # remove ones that are next to each other
    for i in range(1, len(df)):
        if (
            df.loc[i, "transect_num"] == 1
            and df.loc[i - 1, "transect_num"] == 1
            or df.loc[i, "transect_num"] == 1
            and df.loc[i - 2, "transect_num"] == 1
        ):
            df.loc[i, "transect_num"] = 0
    df["transect_num"] = df["transect_num"].shift(-shift)  # recorrect for shift
    df["transect_num"] = df["transect_num"].cumsum() + 1  # 1 indexed
    df["transect_num"] = df["transect_num"].ffill()
    df = df.drop(columns=["prev_azimuth_heading", "deg_displace"])
    return df


def heading_filter(
    df: pd.DataFrame, azimuth_filter: float, azimuth_window: int, elevation_filter: float
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Filters data based on specified azimuth and elevation headings, aiming to isolate transects that
    align with main flight directions. First elevation is filtered to remove significant climbs or descents
    (beware terrain-following flights). Bimodal azimuth headings are calculated and used to filter the
    data based on the main flight directions, with a rolling median applied (the window) to smooth the data.

    Parameters:
        df (pd.DataFrame): The input dataframe.
        azimuth_filter (float): The tolerance for deviation from the main azimuth headings.
        azimuth_window (int): The window size for rolling median calculation of azimuth headings.
        elevation_filter (float): The tolerance for deviation from horizontal flight.

    Returns:
        tuple: The filtered dataframe and the original unfiltered dataframe for comparison.
    """
    df_filtered = df.copy()
    df_filtered = df_filtered[abs(df["elevation_heading"]) < elevation_filter]

    azi1, azi2 = bimodal_azimuth(df_filtered)
    logging.info(f"Drone appears to be flying mainly on the headings {azi1:.2f} degrees and {azi2:.2f} degrees")

    df_filtered["rolling_azimuth_heading"] = (
        df_filtered["azimuth_heading"].rolling(azimuth_window, center=True).apply(lambda x: circ_median(x), raw=True)
    )

    df_filtered = df_filtered[
        (df_filtered["rolling_azimuth_heading"] < azi1 + azimuth_filter)
        & (df_filtered["rolling_azimuth_heading"] > azi1 - azimuth_filter)
        | (df_filtered["rolling_azimuth_heading"] < azi2 + azimuth_filter)
        & (df_filtered["rolling_azimuth_heading"] > azi2 - azimuth_filter)
    ]

    return df_filtered, df


def mCount_max(data_dict: dict[int, float]) -> tuple[int, int]:
    """
    Finds the start and end of the longest monotonic sequence in a dictionary of floats, typically used
    to identify a series of continuous altitude measurements. The first and last are retained.

    Parameters:
    data_dict (Dict[int, float]): Dictionary with sequential numeric keys and numeric values representing measures
    such as altitude.

    Returns:
    tuple: Start and end indices of the longest monotonic sequence in the dictionary.
    """
    if len(data_dict) < 2:
        raise ValueError("Dictionary must contain at least two values")
    if list(data_dict.keys()) != list(range(1, len(data_dict) + 1)):
        raise ValueError("Keys must be sequential integers starting from 1")
    poscount = 0
    negcount = 0
    max_pos_count = 0
    max_pos_transect = 0
    max_neg_count = 0
    max_neg_transect = 0
    pos_start = 0
    neg_start = 0

    for i in range(2, len(data_dict) + 1):
        if data_dict[i] >= data_dict[i - 1]:
            poscount += 1
            negcount = 0
        elif data_dict[i] < data_dict[i - 1]:
            negcount += 1
            poscount = 0

        if max_pos_count < poscount:
            max_pos_count = poscount
            max_pos_transect = i
            pos_start = i - poscount
        elif max_neg_count < negcount:
            max_neg_count = negcount
            max_neg_transect = i
            neg_start = i - negcount

    if max_pos_count > 0 or max_neg_count > 0:
        if max_pos_count >= max_neg_count:
            return pos_start, max_pos_transect
        else:
            return neg_start, max_neg_transect
    else:
        return 0, 0


def largest_monotonic_transect_series(df: pd.DataFrame) -> tuple[pd.DataFrame, int, int]:
    """
    Filters the input dataframe to include only the largest continuous series of transects based on
    monotonic altitude changes.

    Parameters:
        df (pd.DataFrame): The input dataframe with transect and altitude information.

    Returns:
        tuple: The filtered dataframe, start transect, and end transect of the largest monotonic series.
    """
    df = add_transect_azimuth_switches(df)  # heading switches
    alt_dict = dict(df.groupby("transect_num")["altitude_ato"].mean())
    starttransect, endtransect = mCount_max(alt_dict)  # type: ignore
    # filter to the biggest monotonic series of values
    df = df[(df["transect_num"] >= starttransect) & (df["transect_num"] <= endtransect)]
    logging.info(
        f"Parsed a flight of {len(np.unique(df['transect_num']))} transects from {alt_dict[starttransect]:.0f}m"
        f" to {alt_dict[endtransect]:.0f}m between {df['timestamp'].iloc[0]} and {df['timestamp'].iloc[-1]}",
    )
    return df, starttransect, endtransect


def monotonic_transect_groups(df: pd.DataFrame) -> tuple[pd.DataFrame, dict[int, str]]:
    """
    Groups transects into a dict of monotonic transect sequences based on altitude, facilitating analysis of continuous
    flight patterns. Current behaviour is to reuse end transects of previous sequences as the start of the next.

    Parameters:
        df (pd.DataFrame): The input dataframe with transect and altitude information.

    Returns:
        tuple: The dataframe with a new 'group' column indicating the monotonic group ID, and a
        dictionary mapping transects to group IDs.
    """

    df = add_transect_azimuth_switches(df)
    alt_dict = dict(df.groupby("transect")["altitude_ato"].mean())

    group_dict = {}
    previous_altitude = None
    current_group = 1
    previous_trend = None
    first_transect_in_series = True

    for transect, altitude_ato in alt_dict.items():
        if previous_altitude is None:
            group_dict[transect] = f"Group_{current_group}"
        else:
            if altitude_ato == previous_altitude:
                raise ValueError("Error: altitude_ato is the same as the previous transect")
            elif altitude_ato > previous_altitude:
                current_trend = "ascending"
            else:
                current_trend = "descending"

            if current_trend != previous_trend and previous_trend is not None and not first_transect_in_series:
                current_group += 1  # Increment group counter
                first_transect_in_series = True
            if current_trend == previous_trend or previous_trend is None:
                # handles edge case where someone flies up, flies down one transect and then up again (yes really)
                first_transect_in_series = False
            group_dict[transect] = f"Group_{current_group}"
            previous_trend = current_trend
        previous_altitude = altitude_ato
    df["group"] = df["transect_num"].map(group_dict)

    return df, group_dict


def remove_non_transects(
    df: pd.DataFrame,
    chain_length: int = 70,
    azimuth_tolerance: int = 10,
    elevation_tolerance: int = 40,
    smoothing_window: int = 5,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Filters the dataframe to remove segments not matching the criteria for being considered as transects,
    based on azimuth and elevation headings and segment length.

    Parameters:
        df (pd.DataFrame): The input dataframe.
        chain_length (int): Minimum number of consecutive points to be considered a transect. Default is 70.
        azimuth_tolerance (int): Tolerance for deviation from major azimuth headings. Default is 10.
        elevation_tolerance (int): Tolerance for deviation from major elevation headings. Default is 40.
        smoothing_window (int): Window size for rolling median smoothing of headings. Default is 5. Rolling median
            avoids single point errors.

    Returns:
        tuple: Dataframes of removed segments and retained segments that fit transect criteria.
    """

    def get_true_runs(mask):
        enumerated_mask = list(enumerate(mask))  # add index numbers to mask
        # group consecutive true/false values
        groups = groupby(enumerated_mask, key=lambda x: x[1])
        # retain groups of True values
        true_runs = [list(group) for key, group in groups if key]
        return true_runs

    def split_runs_on_azimuth_inversion(df, runs, azimuth_inversion_threshold=120):
        split_runs = []
        for run in runs:
            last_azimuth = df.iloc[run[0][0]]["smoothed_azimuth_heading"]
            current_run = [run[0]]
            for point in run[1:]:
                current_azimuth = df.iloc[point[0]]["smoothed_azimuth_heading"]
                if min_angular_displacement(current_azimuth, last_azimuth) > azimuth_inversion_threshold:
                    split_runs.append(current_run)
                    current_run = [point]
                else:
                    current_run.append(point)
                last_azimuth = current_azimuth
            split_runs.append(current_run)
        return split_runs

    # Create new columns for filtering reasons, initialized with False
    df_removed = df.copy()
    df_removed["filtered_by_azimuth"] = False
    df_removed["filtered_by_elevation"] = False
    df_removed["filtered_by_chain"] = False

    # Apply azimuth filter
    major_azi_headings = bimodal_azimuth(df, heading_col="azimuth_heading")
    major_elev_headings = bimodal_elevation(df, heading_col="elevation_heading")

    # Apply rolling median to azimuth and elevation headings and store them in new columns
    df_removed["smoothed_azimuth_heading"] = (
        df_removed["azimuth_heading"]
        .rolling(smoothing_window, center=True)
        .apply(lambda x: circmean(x, 360, 0), raw=True)
        .fillna(df_removed["azimuth_heading"], inplace=False)
    )

    df_removed["smoothed_elevation_heading"] = (
        df_removed["elevation_heading"]
        .rolling(smoothing_window, center=True)
        .apply(lambda x: circmean(x, 360, 0), raw=True)
        .fillna(df_removed["elevation_heading"], inplace=False)
    )

    azimuth_mask = df_removed["smoothed_azimuth_heading"].apply(
        lambda x: any([min_angular_displacement(x, major) <= azimuth_tolerance for major in major_azi_headings])
    )

    df_removed.loc[~azimuth_mask, "filtered_by_azimuth"] = True

    # Apply elevation filter
    elevation_mask = df_removed["smoothed_elevation_heading"].apply(
        lambda x: any([min_angular_displacement(x, major) <= elevation_tolerance for major in major_elev_headings])
    )
    df_removed.loc[~elevation_mask, "filtered_by_elevation"] = True

    # Combined mask for azimuth and elevation
    mask = azimuth_mask & elevation_mask

    true_runs = get_true_runs(mask)
    split_runs = split_runs_on_azimuth_inversion(df_removed, true_runs)
    chain_mask = [len(run) >= chain_length for run in split_runs]
    chain_filtered_runs = [run for run, filter_by_chain in zip(split_runs, chain_mask, strict=False) if filter_by_chain]
    chain_filtered_indices = [point[0] for run in chain_filtered_runs for point in run]
    df_removed.loc[df_removed.index.difference(chain_filtered_indices), "filtered_by_chain"] = True

    filtered_segments = [df_removed.iloc[run[0][0] : run[-1][0] + 1] for run in chain_filtered_runs]

    df_retained = pd.concat(filtered_segments)
    df_removed = df_removed[~df_removed.index.isin(df_retained.index)]

    return df_removed, df_retained


def flatten_linear_plane(df: pd.DataFrame, distance_filter: float = 10000) -> tuple[pd.DataFrame, float]:
    """
    Transforms a 3D dataset into a linear plane, focusing on the largest contiguous dataset aligned with
    the plane's primary orientation. The new x coordinate is the distance along the plane, y is the distance
    perpendicular to the plane (useful only for deviation), and z is the altitude.

    Parameters:
        df (pd.DataFrame): Input dataframe with "utm_easting", "utm_northing", and "altitude_ato" columns.
        distance_filter (float): Threshold for filtering points based on their distance from the regression line.

    Returns:
        tuple: Modified dataframe with new 'x', 'y', and 'z' columns representing transformed coordinates,
        and the plane's angle of rotation.
    """

    def orthogonal_distance_regression(df: pd.DataFrame) -> tuple[pd.DataFrame, np.ndarray[np.float64, Any]]:
        """
        Perform orthogonal distance regression on the given DataFrame.

        Parameters:
            df (pd.DataFrame): DataFrame containing "utm_easting" and "utm_northing" columns.

        Returns:
            tuple[pd.DataFrame, np.ndarray[np.float64, Any]]: Updated DataFrame with "distance_from_fit" column
            and the fitted parameters (slope, intercept).
        """

        def linear_reg_equation(B, x):
            return B[0] * x + B[1]  # y = mx + c

        required_columns = ["utm_easting", "utm_northing"]
        if not all(col in df.columns for col in required_columns):
            raise ValueError(f"DataFrame must contain columns: {required_columns}")

        model = odr.Model(linear_reg_equation)
        data = odr.Data(df["utm_easting"], df["utm_northing"])

        INITIAL_BETA = [1, 0]  # Initial guess of slope=1, intercept=0
        odr_instance = odr.ODR(data, model, beta0=INITIAL_BETA)
        fit = odr_instance.run()

        if fit.stopreason[0] == "Iteration limit reached":
            raise RuntimeError("ODR fitting failed to converge")

        slope, intercept = fit.beta
        df = df.assign(
            distance_from_fit=abs((slope * df["utm_easting"] - df["utm_northing"] + intercept) / np.sqrt(slope**2 + 1))
        )

        return df, fit.beta

    df, coefs2D = orthogonal_distance_regression(df)
    df = df.loc[df["distance_from_fit"] < distance_filter, :]
    df, coefs2D = orthogonal_distance_regression(df)  # this is intentionally done twice
    df = df.loc[df["distance_from_fit"] < distance_filter, :]
    rotation = np.arctan(coefs2D[0])
    df.loc[:, "x"] = (df["utm_easting"] - df["utm_easting"].min()) * np.cos(-rotation) - (
        df["utm_northing"] - df["utm_northing"].min()
    ) * np.sin(-rotation)
    df.loc[:, "y"] = (df["utm_easting"] - df["utm_easting"].min()) * np.sin(-rotation) + (
        df["utm_northing"] - df["utm_northing"].min()
    ) * np.cos(-rotation)
    df.loc[:, "z"] = df.altitude_ato

    plane_angle = (np.pi / 2) - np.arctan(coefs2D[0])
    return df, plane_angle


## Functions for circular/spiral flights ##


def circle_deviation(df: pd.DataFrame, x_col: str, y_col: str) -> tuple[np.ndarray, np.ndarray, float]:
    """
    Calculates the deviation of points from a fitted circle and their azimuth angles.

    Parameters:
        df (pd.DataFrame): The input dataframe containing the data.
        x (str): Column name for the x-coordinate.
        y (str): Column name for the y-coordinate.

    Returns:
        Tuple[np.ndarray, np.ndarray, float]: A tuple containing the deviations of each point from the circle,
        their azimuth angles, and the radius of the fitted circle.
    """
    required_columns = [x_col, y_col]
    if not all(col in df.columns for col in required_columns):
        raise ValueError(f"DataFrame must contain columns: {required_columns}")

    x = df[x_col]
    y = df[y_col]

    def midhalf(x):  # select middle half of data to avoid edge effects
        return x.iloc[int(len(x) * 1 / 4) : int(len(x) * 3 / 4)]

    x_filter, y_filter = midhalf(x), midhalf(y)

    def func(params):
        xc, yc, r = params
        return np.sqrt((x - xc) ** 2 + (y - yc) ** 2) - r

    x_m = np.mean(np.array(x_filter))  # initial guess for parameters
    y_m = np.mean(np.array(y_filter))
    r_m = np.mean(np.sqrt((np.array(x_filter) - x_m) ** 2 + (np.array(y_filter) - y_m) ** 2))

    params0 = np.array([x_m, y_m, r_m])
    result = least_squares(func, params0)
    xc, yc, r = result.x

    deviation = np.sqrt((x - xc) ** 2 + (y - yc) ** 2) - r

    # output azimuth in radians with 0 at north and increasing clockwise thanks to modulos
    azimuth = np.degrees(np.arctan2(x - xc, y - yc) % (2 * np.pi))

    return deviation, azimuth, r


def recentre_azimuth(df: pd.DataFrame, r: float, x: str = "circ_azimuth", y: str = "ch4_normalised") -> pd.DataFrame:
    """
    Recentres the azimuth angles based on the angle of maximum value and computes the distance along the
    circumference of the circle.

    Parameters:
        df (pd.DataFrame): The input dataframe.
        r (float): The radius of the circle.
        x (str): Column name for azimuth angles. Default is 'circ_azimuth'.
        y (str): Column name for the values used to find the maximum. Default is 'ch4_normalised'.

    Returns:
        pd.DataFrame: The modified dataframe with centered azimuth angles and distances along the circumference.
    """

    def azimuth_of_max(df: pd.DataFrame, x: str = "circ_azimuth", y: str = "ch4_normalised") -> float:
        return df.loc[df[y].idxmax()][x]

    centre_azimuth = azimuth_of_max(df, x, y)
    df["centred_azimuth"] = df[x] - centre_azimuth
    df.loc[df["centred_azimuth"] > 180, "centred_azimuth"] -= 360
    df.loc[df["centred_azimuth"] < -180, "centred_azimuth"] += 360
    df["circumference_distance"] = r * np.radians(df["centred_azimuth"])
    df["circumference_distance"] = df["circumference_distance"] - df["circumference_distance"].min()
    return df


def wind_rel_ground(
    df: pd.DataFrame, aircraft_u: np.ndarray, aircraft_v: np.ndarray, wind_u: np.ndarray, wind_v: np.ndarray
) -> pd.DataFrame:
    """
    Calculates wind speed and direction relative to the ground from aircraft and airspeed vectors.
    u = N, v = E

    Parameters:
        df (pd.DataFrame): The input dataframe.
        aircraft_u (np.ndarray): The northward component of the aircraft's velocity.
        aircraft_v (np.ndarray): The eastward component of the aircraft's velocity.
        wind_u (np.ndarray): The northward component of the wind's velocity.
        wind_v (np.ndarray): The eastward component of the wind's velocity.

    Returns:
        pd.DataFrame: The dataframe updated with ground-relative wind speed and direction.
    """
    u_wind_ground = wind_u - aircraft_u
    v_wind_ground = wind_v - aircraft_v
    windspeed_ground = np.sqrt(u_wind_ground**2 + v_wind_ground**2)
    winddir_ground = np.degrees(np.arctan2(u_wind_ground, v_wind_ground) % (2 * np.pi))
    df["windspeed_ground"] = windspeed_ground
    df["winddir_ground"] = winddir_ground
    return df
